/*
 * warsow.onInit( void() cb )
 * warsow.onMapSpawn( void() cb )
 * warsow.onShutdown( void() cb )
 * warsow.onStateStarted( void() cb )
 * warsow.onStateIncoming( bool( MatchState incoming ) cb )
 * warsow.onThink( void() cb )
 * warsow.onRespawn( Entity @ent, Team old, Team new )
 * warsow.onScore( Client @client, ScoreEvent event, String &args )
 *
 * warsow.command( String &command, bool( Client @client, String &args ) cb )
 * warsow.callvote( String &vote, bool( ... ) validate, bool( ... ) passed )
 *
 * warsow.roundStart()
 * warsow.roundEnd()
 * warsow.onRoundStart( void() cb )
 * warsow.onRoundEnd( void() cb )
 */

enum score_e {
	SCORE_ERROR,
	SCORE_CONNECT,
	SCORE_ENTERGAME,
	SCORE_DISCONNECT,
	SCORE_DAMAGE,
	SCORE_KILL,
	SCORE_PROJECTILEHIT,
	SCORE_USERINFOCHANGED,
	SCORE_AWARD,
	SCORE_PICKUP,
};

enum roundstates_e {
	ROUNDSTATE_NONE,
	ROUNDSTATE_PRE,
	ROUNDSTATE_PLAYING,
	ROUNDSTATE_GRACE,
	ROUNDSTATE_POST,
};

funcdef void void_cb();
funcdef bool matchstate_cb( const matchstates_e );
funcdef void respawn_cb( Entity @, const teams_e, const teams_e );
funcdef void score_cb( Client @, const score_e, const String & );
funcdef bool roundstate_cb( const roundstates_e );

funcdef bool command_cb( Client @, const String &, const uint );

/* awards:
 *   most damage
 *   most efficient
 *   most accurate
 *   highest top speed?
 *   most distance travelled?
 *   bongo/impressive
 */

class RoundStats {
	uint damage;
	uint teamDamage;
	uint kills;
	uint shotsFired;
	uint shotsHit;
}

class Warsow {
	// array< void_cb @ > cbsInit;
	array< void_cb @ > cbsMapSpawn;
	array< void_cb @ > cbsShutdown;
	array< void_cb @ > cbsStateStarted;
	array< matchstate_cb @ > cbsStateIncoming;
	array< void_cb @ > cbsThink;
	array< respawn_cb @ > cbsRespawn;
	array< score_cb @ > cbsScore;

	array< void_cb @ > cbsRoundStart;
	array< void_cb @ > cbsRoundEnd;
	array< void_cb @ > cbsRoundStateStarted;
	array< roundstate_cb @ > cbsRoundStateIncoming;

	Dictionary score_events;

	array< String > cmdsCommands;
	array< command_cb @ > cmdsCallbacks;

	RoundStats[] stats( maxClients );

	Warsow() {
		this.score_events.set( "connect", SCORE_CONNECT );
		this.score_events.set( "enterGame", SCORE_ENTERGAME );
		this.score_events.set( "disconnect", SCORE_DISCONNECT );
		this.score_events.set( "dmg", SCORE_DAMAGE );
		this.score_events.set( "kill", SCORE_KILL );
		this.score_events.set( "projectilehit", SCORE_PROJECTILEHIT );
		this.score_events.set( "userinfochanged", SCORE_USERINFOCHANGED );
		this.score_events.set( "award", SCORE_AWARD );
		this.score_events.set( "pickup", SCORE_PICKUP );

		this.onThink( GENERIC_Think );
		this.onScore( GENERIC_Awards );

		this.command( "gametype", GENERIC_CommandGametype );
		this.command( "cvarinfo", GENERIC_CommandCvarinfo );

		this.onRoundEnd( GENERIC_RoundAwards );
	}

	// void onInit( const void_cb @ cb ) {
	// 	this.cbsInit.insertLast( cb );
	// }

	void onMapSpawn( const void_cb @ cb ) {
		this.cbsMapSpawn.insertLast( cb );
	}

	void onShutdown( const void_cb @ cb ) {
		this.cbsShutdown.insertLast( cb );
	}

	void onStateStarted( const void_cb @ cb ) {
		this.cbsStateStarted.insertLast( cb );
	}

	void onStateIncoming( const matchstate_cb @ cb ) {
		this.cbsStateIncoming.insertLast( cb );
	}

	void onThink( const void_cb @ cb ) {
		this.cbsThink.insertLast( cb );
	}

	void onRespawn( const respawn_cb @ cb ) {
		this.cbsRespawn.insertLast( cb );
	}

	void onScore( const score_cb @ cb ) {
		this.cbsScore.insertLast( cb );
	}

	void onRoundStart( const void_cb @ cb ) {
		this.cbsRoundStart.insertLast( cb );
	}

	void onRoundEnd( const void_cb @ cb ) {
		this.cbsRoundEnd.insertLast( cb );
	}

	void onRoundStateStarted( const void_cb @ cb ) {
		this.cbsRoundStateStarted.insertLast( cb );
	}

	void onRoundStateIncoming( const roundstate_cb @ cb ) {
		this.cbsRoundStateIncoming.insertLast( cb );
	}

	void command( const String & command, const command_cb @ cb ) {
		int idx = this.cmdsCommands.find( command );

		if( idx != -1 ) {
			G_Print( "ERROR: Command `" + command + "` already exists.\n" );
		}
		else {
			this.cmdsCommands.insertLast( command );
			this.cmdsCallbacks.insertLast( cb );

			G_RegisterCommand( command );
		}
	}
}

Warsow warsow;

// void GT_InitGametype() {
// 	for( uint i = 0; i < warsow.cbsInit.size(); i++ ) {
// 		warsow.cbsInit[ i ]();
// 	}
// }

void GT_SpawnGametype() {
	for( uint i = 0; i < warsow.cbsMapSpawn.size(); i++ ) {
		warsow.cbsMapSpawn[ i ]();
	}
}

void GT_Shutdown() {
	for( uint i = 0; i < warsow.cbsShutdown.size(); i++ ) {
		warsow.cbsShutdown[ i ]();
	}
}

void GT_MatchStateStarted() {
	for( uint i = 0; i < warsow.cbsStateStarted.size(); i++ ) {
		warsow.cbsStateStarted[ i ]();
	}
}

bool GT_MatchStateFinished( int incomingMatchState ) {
	bool advance = true;

	for( uint i = 0; i < warsow.cbsStateIncoming.size(); i++ ) {
		advance = advance && warsow.cbsStateIncoming[ i ]( matchstates_e( incomingMatchState ) );
	}

	return advance;
}

void GT_ThinkRules() {
	for( uint i = 0; i < warsow.cbsThink.size(); i++ ) {
		warsow.cbsThink[ i ]();
	}
}

void GT_PlayerRespawn( Entity @ ent, int old_team, int new_team ) {
	for( uint i = 0; i < warsow.cbsRespawn.size(); i++ ) {
		warsow.cbsRespawn[ i ]( ent, teams_e( old_team ), teams_e( new_team ) );
	}
}

void GT_ScoreEvent( Client @ client, const String & score_event, const String & args ) {
	int64 event = SCORE_ERROR;
	bool ok = warsow.score_events.get( score_event, event );

	if( !ok ) {
		G_Print( "ERROR: Unrecognised score event: " + score_event + "\n" );
	}

 	for( uint i = 0; i < warsow.cbsScore.size(); i++ ) {
		warsow.cbsScore[ i ]( client, score_e( event ), args );
 	}
}

bool GT_Command( Client @ client, const String & command, const String & args, int argc ) {
	int idx = warsow.cmdsCommands.find( command );

	if( idx == -1 ) {
		return false;
	}

	return warsow.cmdsCallbacks[ idx ]( client, args, uint( argc ) );
}

void GENERIC_Awards( Client @ client, const score_e score_event, const String & args ) {
	if( score_event == SCORE_KILL ) {
		Entity @ attacker = client is null ? null : client.getEnt();
		Entity @ victim = G_GetEntity( args.getToken( 0 ).toInt() );
		Entity @ inflictor = G_GetEntity( args.getToken( 1 ).toInt() );

		award_playerKilled( victim, attacker, inflictor );
	}
}

void GENERIC_RoundAwards() {
}

bool GENERIC_CommandGametype( Client @ client, const String & args, const uint argc ) {
	String response = "";
	Cvar fs_game( "fs_game", "", 0 );
	String manifest = gametype.manifest;

	response += "\n";
	response += "Gametype " + gametype.name + " : " + gametype.title + "\n";
	response += "----------------\n";
	response += "Version: " + gametype.version + "\n";
	response += "Author: " + gametype.author + "\n";
	response += "Mod: " + fs_game.string + ( !manifest.empty() ? " (manifest: " + manifest + ")" : "" ) + "\n";
	response += "----------------\n";

	G_PrintMsg( client.getEnt(), response );

	return true;
}

bool GENERIC_CommandCvarinfo( Client @ client, const String & args, const uint argc ) {
	GENERIC_CheatVarResponse( client, "", args, argc );

	return true;
}
